/*********************************************************************

   Copyright (C) 2002 Smaller Animals Software, Inc.

   This software is provided 'as-is', without any express or implied
   warranty.  In no event will the authors be held liable for any damages
   arising from the use of this software.

   Permission is granted to anyone to use this software for any purpose,
   including commercial applications, and to alter it and redistribute it
   freely, subject to the following restrictions:

   1. The origin of this software must not be misrepresented; you must not
   claim that you wrote the original software. If you use this software
   in a product, an acknowledgment in the product documentation would be
   appreciated but is not required.

   2. Altered source versions must be plainly marked as such, and must not be
   misrepresented as being the original software.

   3. This notice may not be removed or altered from any source distribution.

   http://www.smalleranimals.com
   smallest@smalleranimals.com
			  
**********************************************************************/

/**********************************************************************

  CSAScriptEval is the main parsing object.
  
**********************************************************************/

#include "stdafx.h" 
#include "SAScrDefs.h"
#include "SAScrEval.h"
#include "SAScrEvalFuncDefs.h" 
#include "SAScrUtils.h"           
#include "SAScrException.h" 
//////////////////////////////////////////////////////////////////////////

// special chars for this grammar
TCHAR CSAScriptEval::m_MetaChars[] = 
{
	g_operSeparatorStart, 
		g_operSeparatorStop, 
		g_operParamSeparator, 
		g_variablePrefix, 
		_T( '"' ),                    // literal text is always in quotes
};
int  CSAScriptEval::m_iMetaChars = ARRAYSIZE(m_MetaChars);

#define MACRO_DEFINITION_START _T( "Macro:" )

//////////////////////////////////////////////////////////////////////////
// initialize the evaluator with a map of variables

CSAScriptEval::CSAScriptEval(const CSAScrVariableMap &vars) : 
m_tokenizer(NULL, 0, m_MetaChars, m_iMetaChars) // no keywords, just metachars
{
	// you must initialize the global var oper manager
	assert(g_varOperManager.IsInit());
	
	// copy the vars
	m_localVarMap = vars;
	
#ifndef SASCR_SILENT
	m_localVarMap.Dump();
#endif
	
	// build a little string
	m_csSepString = _T( "" );
	m_csSepString += g_operSeparatorStart;
	m_csSepString += g_operSeparatorStop;
	m_csSepString += g_operParamSeparator;
}

//////////////////////////////////////////////////////////////////////////

// protected constructor
CSAScriptEval::CSAScriptEval(const CSAScrVariableMap &localVars, 
                             const CSAScrVariableMap &userVars): 
m_tokenizer(NULL, 0, m_MetaChars, m_iMetaChars) // no keywords, just metachars
{
	// you must initialize the global var oper manager
	assert(g_varOperManager.IsInit());
	
	// copy the vars
	m_localVarMap = localVars;
	
#ifndef SASCR_SILENT
	m_localVarMap.Dump();
#endif
	
	// copy these vars
	m_userVarMap = userVars;
	
#ifndef SASCR_SILENT
	m_userVarMap.Dump();
#endif
	
	// build a little string
	m_csSepString = _T( "" );
	m_csSepString += g_operSeparatorStart;
	m_csSepString += g_operSeparatorStop;
	m_csSepString += g_operParamSeparator;
}

/******************************************************

  CSAScriptEval::ParseString
  
	CSAScriptEval main entry point. 
	
	  csText   expression to evaluate
	  eCtx     variable context (usually CTXALL)
	  
		returns evalatued string - with possible error text
		
******************************************************/
CSAString CSAScriptEval::ParseString(const TCHAR	* pText,
									 eParseContext eCtx)
									 
{
	CSAString csOut;
	try
	{
		m_eCtx = eCtx;
		
		m_iRefCount = 0;
		
		ParseStringA(csOut, pText);
	}
	catch (CSAScrException eError)
	{
		//assert(0);
		csOut = eError.Text();
	}
	catch (...)
	{
		//assert(0);
		csOut=CSAScrException::TranslateError(CSAScrException::eParseInternal);
	}
	
	return csOut;
}

//////////////////////////////////////////////////////////////////////////
// look for _T( "&" ) . parse after it, look for the next _T( "&" )...

void CSAScriptEval::ParseStringA(CSAString & csOutText, const CSAString &csText)
{
	
	try
	{
		csOutText = _T( "" );
		
		if (m_iRefCount > VARS_MAX_REF)
		{
			throw CSAScrException(CSAScrException::eParseRecursionLimit, _T( "" ));
		}
		
		if (m_tokenizer.Tokenize(csText)!=CTokenizer::eErrorNone)
		{
			throw CSAScrException(CSAScrException::eParseSyntax, m_tokenizer.GetLastToken().m_csToken);
		}
		
#ifndef SASCR_SILENT
		m_tokenizer.Dump();
#endif
		
		m_iRefCount++;
		
		int iLastUsedExpStop = 0;
		
		int iTokens = m_tokenizer.GetTokenCount();
		if (iTokens > 0)
		{
			for (int i=0; i < iTokens; i++)
			{
				CSAToken curToken = m_tokenizer.GetToken(i);
				
				// looking for expression start...
				if (curToken.m_csToken.GetLength()==1)
				{
					if (curToken.m_csToken.GetAt(0)==g_variablePrefix && !curToken.m_bIsQuotedString)
					{
						// starting an expression... maybe
						if (i < iTokens - 1)
						{
							// skip the expression start token
							i++;
							
							if (!IsExpression(m_tokenizer, i))
							{
								// nothing to do here
								continue;
							}
							
							// output from iLastUsedExpStop to this token
							// token.m_iStop - token.m_iStart might not always equal token.m_csToken.GetLength
							// so, we need to backtrack from curToken.m_iStop to find the real start of the
							// token text
							int iStart = iLastUsedExpStop;
							int iStop = curToken.m_iStop - curToken.m_csToken.GetLength() + 1; 
							int iLen = iStop - iStart;
							assert(iLen >= 0);
							
							iLen = max(0, iLen);
							csOutText += csText.Mid(iStart, iLen);
							
							// how long is this expression?
							int iEndToken = FindLastToken(m_tokenizer, i);
							
							if (iEndToken < 0)
							{
								if (iTokens > i - 1)
								{
									throw CSAScrException(CSAScrException::eParseSyntaxNonterm, curToken.m_csToken + m_tokenizer.GetToken(i+1).m_csToken);
								}
								
								throw CSAScrException(CSAScrException::eParseSyntaxNonterm, _T( "(no operator)" ));
							}
							
							// do it
							csOutText += Evaluate(i, iEndToken);
							
							// update
							CSAToken endToken = m_tokenizer.GetToken(iEndToken);
							iLastUsedExpStop = endToken.m_iStop + 1;
							
							i = iEndToken;
							
							continue;
							
						} // not last token
						
					} // token = exp start
					
				} // token len = 1
				
			} // for
			
			// output from iLastExpStop to this token
			csOutText += csText.Right(csText.GetLength() - iLastUsedExpStop);
			
		} // no tokens
	}
	catch (CSAScrException eError)
	{
		//assert(0);
		csOutText+=eError.Text();
	}
	catch (...)
	{
		//assert(0);
		csOutText+=CSAScrException::TranslateError(CSAScrException::eParseInternal);
	}
}

//////////////////////////////////////////////////////////////////////////

void CSAScriptEval::FindExpressions(vector<COLORREF> & vOutColors, 
                                    const CSAString &csText,
                                    COLORREF clrNon,
                                    COLORREF clrExp,
                                    COLORREF clrQot,
                                    COLORREF clrOper,
                                    COLORREF clrVar,
                                    COLORREF clrErr) const
{
	CSAToken curToken;
	CSAToken endToken;
	
	try
	{
		// clear
		vOutColors.resize(csText.GetLength());
		for (vector<COLORREF>::iterator it = vOutColors.begin(); it!=vOutColors.end(); it++)
		{
			(*it) = clrNon;
		}
		
		CTokenizer      tokenizer(NULL, 0, m_MetaChars, m_iMetaChars); // no keywords, just metachars
		
		if (tokenizer.Tokenize(csText)!=CTokenizer::eErrorNone)
		{
			throw CSAScrException(CSAScrException::eParseSyntax, tokenizer.GetLastToken().m_csToken);
		}
		
		int iLastUsedExpStop = 0;
		
		int iTokens = tokenizer.GetTokenCount();
		if (iTokens > 0)
		{
			for (int i=0; i < iTokens; i++)
			{
				curToken = tokenizer.GetToken(i);
				
				// looking for expression start...
				if (curToken.m_csToken.GetLength()==1)
				{
					if (curToken.m_csToken.GetAt(0)==g_variablePrefix && !curToken.m_bIsQuotedString)
					{
						// starting an expression... maybe
						if (i < iTokens - 1)
						{
							// skip the expression start token
							i++;
							
							if (!IsExpression(tokenizer, i))
							{
								// nothing to do here
								continue;
							}
							
							// output from iLastUsedExpStop to this token
							// token.m_iStop - token.m_iStart might not always equal token.m_csToken.GetLength
							// so, we need to backtrack from curToken.m_iStop to find the real start of the
							// token text
							int iStart = iLastUsedExpStop;
							int iStop = curToken.m_iStop - curToken.m_csToken.GetLength() + 1; 
							int iLen = iStop - iStart;
							assert(iLen >= 0);
							
							iLen = max(0, iLen);
							//							csOutText += csText.Mid(iStart, iLen);
							
							// how long is this expression?
							int iEndToken = FindLastToken(tokenizer, i);
							
							if (iEndToken < 0)
							{
								if (iTokens > i - 1)
								{
									throw CSAScrException(CSAScrException::eParseSyntaxNonterm, curToken.m_csToken + tokenizer.GetToken(i+1).m_csToken);
								}
								
								throw CSAScrException(CSAScrException::eParseSyntaxNonterm, _T( "(no operator)" ));
							}
							
							// update
							endToken = tokenizer.GetToken(iEndToken);
							
							// set the colors (-1, because i has already been incd. <= because we count the end token)
							for (int iExpTokenIdx = i - 1; iExpTokenIdx <= iEndToken; iExpTokenIdx++)
							{
								CSAToken cur = tokenizer.GetToken(iExpTokenIdx);
								bool bIsVar = IsVariable(cur.m_csToken);
								
								CSAScrOperator dummyOper;
								bool bIsOper = GetOperator(cur.m_csToken, dummyOper);
								
								//
								//	[JRT] Changed _T( "int iChar" ) To _T( "DWORD dwChar" ): STL
								//	Objects Tend To Use _T( "unsigned int" ) For Sizes And 
								//	Indices, Also reduces To One igned/unisgned 
								//	Mismatch, In The Conditional Of The _T( "for" ) Loop Below.
								//
								//for (int iChar = cur.m_iStart; iChar <= cur.m_iStop; iChar++)
								for (DWORD dwChar = cur.m_iStart; dwChar <= (DWORD)cur.m_iStop; dwChar++)
								{
									assert(dwChar < vOutColors.size());
									if (cur.m_bIsQuotedString)
									{
										vOutColors.at(dwChar) = clrQot;
									}
									else if (bIsVar)
									{
										vOutColors.at(dwChar) = clrVar;
									}
									else if (bIsOper)
									{
										vOutColors.at(dwChar) = clrOper;
									}
									else
									{
										vOutColors.at(dwChar) = clrExp;                              
									}
								}
							}
							
							// get the last paren
							
							
							iLastUsedExpStop = endToken.m_iStop + 1;
							i = iEndToken;
							
							continue;
							
						} // not last token
						
					} // token = exp start
					
				} // token len = 1
				
			} // for
			
		} // no tokens
	}
	catch (...)
	{
		int iEnd = csText.GetLength();
		for (int iChar = curToken.m_iStart; iChar < iEnd; iChar++)
		{
			vOutColors.at(iChar) = clrErr;
		}
	}
}

//////////////////////////////////////////////////////////////////////////
// find the last ) for this particular expression

int CSAScriptEval::FindLastToken(CTokenizer &tokenizer, int iStartToken) const
{
	int iFirstOpen = -1;
	
	// find first open paren
	int iTokens = tokenizer.GetTokenCount();
	for (int i = iStartToken; i < iTokens ;i++)
	{
		CSAToken cur = tokenizer.GetToken(i);
		if (cur.m_csToken.GetLength()==1) // len = 1
		{
			if (cur.m_csToken.GetAt(0) == g_operSeparatorStart && !cur.m_bIsQuotedString)
			{
				iFirstOpen = i;
				break;
			}
		}
	}
	
	if (iFirstOpen==-1)
	{
		return -1;
	}
	
	// now, check balance to find the matching close paren
	int iOpen = 1;
	int iClose = 0;
	
	for (i = iFirstOpen + 1; i < iTokens; i++)
	{
		CSAToken cur = tokenizer.GetToken(i);
		
		if (cur.m_csToken.GetLength()==1) // len = 1
		{
			if (cur.m_csToken.GetAt(0)==g_operSeparatorStop && !cur.m_bIsQuotedString)
			{
				iClose++;
			}
			
			if (cur.m_csToken.GetAt(0)==g_operSeparatorStart && !cur.m_bIsQuotedString)
			{
				iClose--;
			}
		}
		
		if (iOpen==iClose)
		{
			// gotcha
			return i;
		}
	}
	
	return -1;
}

//////////////////////////////////////////////////////////////////////////
// has an operator just before an open paren

bool CSAScriptEval::IsExpression(CTokenizer &tokenizer, int iStartToken) const
{
	bool bIsExp = false;
	
	// cur token must be an operator
	CSAToken curToken = tokenizer.GetToken(iStartToken);
	
	CSAScrOperator curOper;
	if (GetOperator(curToken.m_csToken, curOper))
	{
		if (iStartToken + 1 < tokenizer.GetTokenCount())
		{
			// next token must be an open paren
			curToken = tokenizer.GetToken(iStartToken + 1);
			if (curToken.m_csToken.GetLength()==1) // len = 1
			{
				bIsExp = (curToken.m_csToken.GetAt(0) == g_operSeparatorStart && !curToken.m_bIsQuotedString);
			}
		}
	}
	
	return bIsExp;
}

//////////////////////////////////////////////////////////////////////////
// main eval loop

CSAString CSAScriptEval::Evaluate(int iStartToken, int iEndToken)
{
	if (iStartToken > iEndToken)
	{
		throw CSAScrException(CSAScrException::eParseInternal, _T( "No token" ));
	}
	
	CSAString csOut;
	
	CSAScrOpStack expressionStack;
	CSAScrOpStack operandStack;
	
	// stack-ize the expression in m_tokenizer from iStartToken to iEndToken. 
	ExpressionToStack(expressionStack, m_tokenizer, iStartToken, iEndToken);
	
	bool done = false;
	try
	{
		while (!done)
		{
			if (!expressionStack.empty())
			{
				// pop!
				CSAScrOp opTop = expressionStack.top();
				expressionStack.pop();
				
				// do we have an operator ? 
				// ExpressionToStack decides operator m_vs operand
				switch (opTop.m_type)
				{
				case CSAScrOp::eOperator:
					{
						// is this an operator?
						CSAScrOperator cur = opTop.m_operator;
						
						// sanity check
						if (cur.m_pFnOperator!=NULL)
						{
							CSAScrOpStack currentParamStack;
							
							// grab the correct number of operands (parameters)
							for (int i=0; i < cur.m_iParamCount; i++)
							{
								if (!operandStack.empty())
								{
									// from here
									CSAScrOp param = operandStack.top();
									operandStack.pop();
									
									// to here
									currentParamStack.push(param);
								}
								else
								{
									CSAString t;
									t.Format(_T( "Operator %s expects %d parameters" ), cur.m_csOperatorName, cur.m_iParamCount);
									throw CSAScrException(CSAScrException::eParseParameterCount, t);
								}
							}
							
							// evaluate
							CSAString csRes = DoOperator(cur, currentParamStack);
							
							// stash it away
							expressionStack.push(CSAScrOp(csRes));
						}
						else
						{
							SAScr::SAScrTrace(_T( "*** NULL operator\n" ));
							throw CSAScrException(CSAScrException::eParseInternal, _T( "NULL operator" ));
						}
					}
					break;
				case CSAScrOp::eOperand:
					{
						// if the expression stack is empty this must be the final result.
						if (expressionStack.empty())
						{
							csOut = opTop.m_text;
							done = true;
						}
						else
						{
							// save this for the next operator
							operandStack.push(opTop);
						}
					}
					break;
				default:
					SAScr::SAScrTrace(_T( "*** Unknown CSAScrOp type %d\n" ), (int)opTop.m_type);
					throw CSAScrException(CSAScrException::eParseInternal, _T( "Unknown symbol type on stack" ));
					break;
				}
				
			}
			else
			{
				done = true;
			}
			
		} // while !done
		
		
		if (!operandStack.empty())
		{
#ifndef SASCR_SILENT
			SAScr::SAScrTrace(_T( "Operands left on the stack\n" ));
			while (!operandStack.empty())
			{
				CSAScrOp param = operandStack.top();
				operandStack.pop();
				SAScr::SAScrTrace(_T( "%s\n" ), param.m_text);
			}
#endif
			throw CSAScrException(CSAScrException::eParseParameterCount, _T( "" ));
		}
		
		
	} // try
	catch (CSAScrException eError)
	{
		//assert(0);
		csOut+=eError.Text();
	}
	catch (...)
	{
		//assert(0);
		csOut+=CSAScrException::TranslateError(CSAScrException::eParseInternal);
	}
	return csOut;
}

//////////////////////////////////////////////////////////////////////////
// decompose the expression in m_tokenizer, from iStartToken to iEndToken, 
// onto a stack.

void CSAScriptEval::ExpressionToStack(CSAScrOpStack &expStack, 
                                      CTokenizer &tokenizer,
									           int iStartToken,
									           int iEndToken)
{
	CSAString csToken, csNextToken;
	
   enum {eStart, eOperator, eVariable, eString, eNumeric, eMetachar, eNone} eCurType;
   eCurType = eStart;

	CSAScrOperator curOper;
	
	if (iStartToken < 0)
	{
		throw CSAScrException(CSAScrException::eParseInternal, _T("Bad token position"));
	}
	
	CSAToken curToken, nextToken;
	curToken = tokenizer.GetToken(iStartToken);
	csToken = curToken.m_csToken;
	
	for (int i=iStartToken; i < iEndToken; i++)
	{
      // reset
      eCurType = eNone;

		if (i > iStartToken)
		{
			curToken = nextToken;
			csToken = curToken.m_csToken;
		}
		
		if (i < iEndToken - 1)
		{
			nextToken = tokenizer.GetToken(i + 1);
			csNextToken = nextToken.m_csToken;
		}
		
		/////////////////////////////////////////////////////////////
		// handle empty parameters
		if ((csToken.GetLength()==1 && 
         !curToken.m_bIsQuotedString && 
         csNextToken.GetLength()==1 && 
         !nextToken.m_bIsQuotedString))
		{
			// (,
			if ((csToken.GetAt(0)==g_operSeparatorStart) && (csNextToken.GetAt(0)==g_operParamSeparator))
			{
				csToken = "";
			}
			
			// ,,
			else if ((csToken.GetAt(0)==g_operParamSeparator) && (csNextToken.GetAt(0)==g_operParamSeparator))
			{
				csToken = "";
			}
			
			// ,)
			else if ((csToken.GetAt(0)==g_operParamSeparator) && (csNextToken.GetAt(0)==g_operSeparatorStop))
			{
				csToken = "";
			}
		}

		/////////////////////////////////////////////////////////////

		bool bNextIsParen = false;
		if (csNextToken.GetLength()==1 && !nextToken.m_bIsQuotedString)
		{
			bNextIsParen = (csNextToken.GetAt(0)==g_operSeparatorStart);
		}
      
		/////////////////////////////////////////////////////////////
		// skip separators, variables and operators
		if (csToken.IsOneOf(m_csSepString) && !curToken.m_bIsQuotedString)
		{
         eCurType = eMetachar;
		}
		else if (csToken.IsOneOf(&g_variablePrefix))
		{
			eCurType = eOperator;
		}
      // quoted strings are always operands
      else if (curToken.m_bIsQuotedString)
		{
			// quoted text
         csToken = tokenizer.TranslateEscapes(csToken);

			expStack.push(CSAScrOp(csToken));

         eCurType = eString;
		}
      // un-quoted string. must be a number, a variable, or an operator
		else if (IsVariable(csToken)) // is it a variable?
		{
			// variable? expand it
			csToken = ParseVariable(csToken);
			
			// operand 
			expStack.push(CSAScrOp(csToken));
			
			eCurType = eVariable;
		} 
      // numeric operand ?
		else if (SAScr::IsAllDigits(csToken, true)) 
		{
			// yes. it's a number. push it.
			expStack.push(CSAScrOp(csToken));
			
         eCurType = eNumeric;
		}
      // is this an operator?
		else if (GetOperator(csToken, curOper)) 
		{
         if (bNextIsParen)
         {
            // is it a macro ?
            if (curOper.m_bMacroFunction)
            {
               // stop the presses!
               // we need to expand this macro before we go any farther
               int iMacroEnd = FindLastToken(tokenizer, i);
               if (iMacroEnd >= 0)
               {
                  // this will parse and expand the macro onto the current expStack
                  ExpandMacro(curOper, tokenizer, i, iMacroEnd + 1, expStack);

                  // skip all tokens associated with the macro
                  i = iMacroEnd + 1;
		            if (i < iEndToken - 1)
		            {
			            nextToken = tokenizer.GetToken(i + 1);
			            csNextToken = nextToken.m_csToken;
		            }

                  eCurType = eOperator;
               }
               else
               {
                  throw CSAScrException(CSAScrException::eParseSyntaxNonterm, csToken);
               }
            }
            else
            {
			      // it's a normal operator
			      expStack.push(CSAScrOp(curOper));

               eCurType = eOperator;
            }
         }
         else
         {
            CSAString t;
            t.Format(_T("Operator %s missing parameter list"), curOper.m_csOperatorName);
            throw CSAScrException(CSAScrException::eParseSyntax, t);
         }
      }

      // a little syntax checking
      switch (eCurType)
      {
      case eOperator:
      case eMetachar:
         continue;
         break;
      case eStart:
      case eVariable:
      case eString:
      case eNumeric:
         // operators always precede an open paren, but nothing else should
         if (bNextIsParen)
         {
            CSAString t;
            t.Format(_T("%s is not a valid operator"), curToken.m_csToken);
            throw CSAScrException(CSAScrException::eParseSyntax, t);
         }
         continue;
         break;
      case eNone:
      default:
         break;
      }
      
		// unknown symbol (unknown or out of context variable, bad operator name, etc.)
#ifdef _DEBUG
      TRACE("unknown symbol: %s\n", csToken);
#endif
      throw CSAScrException(CSAScrException::eParseUnknownSymbol, csToken);
	}
}


//////////////////////////////////////////////////////////////////////////
// is csText in our global or local variable maps?

bool CSAScriptEval::IsVariable(const CSAString& csText) const
{
	// first, try global map
	
	// variable names are stored in the map as upper case
	CSAString csUpperVar = csText;
	csUpperVar.MakeUpper();
	
	// look up
	CSAScrVariableMap::const_iterator theIterator;
	theIterator = g_varOperManager.m_variableMap.find(csUpperVar);
	
	// did we get one?
	if (theIterator!=g_varOperManager.m_variableMap.end())
	{
		// only true if we allow this variable in this context
		return IsCTX((*theIterator).second.dwContext, CTX(m_eCtx));
	}
	
	// now try the local map
	theIterator = m_localVarMap.find(csUpperVar);
	
	// did we get one?
	if (theIterator!=m_localVarMap.end())
	{
		// only true if we allow this variable in this context
		return IsCTX((*theIterator).second.dwContext, CTX(m_eCtx));
	}
	
	// now try the user map
	theIterator = m_userVarMap.find(csUpperVar);
	
	// did we get one?
	if (theIterator!=m_userVarMap.end())
	{
		// context doesn't apply here
		return true;
	}
	
	// not found
	return false;
}

//////////////////////////////////////////////////////////////////////////
// get the text for the variable

CSAString CSAScriptEval::ParseVariable(const CSAString &csText)
{
	// first, try global map
	
	// variable names are stored in the map as upper case
	CSAString csUpperVar = csText;
	csUpperVar.MakeUpper();
	
	// look up
	CSAScrVariableMap::iterator theIterator;
	theIterator = g_varOperManager.m_variableMap.find(csUpperVar);
	
	// did we get one?
	if (theIterator!=g_varOperManager.m_variableMap.end())
	{
		// only evaluate if this variable is valid in this context
		if (IsCTX((*theIterator).second.dwContext, CTX(m_eCtx)))
		{
			return (*theIterator).second.pValue;
		}
		else
		{
			return csText;
		}
	}
	
	// now try the local map
	theIterator = m_localVarMap.find(csUpperVar);
	
	// did we get one?
	if (theIterator!=m_localVarMap.end())
	{
		// only evaluate if this variable is valid in this context
		if (IsCTX((*theIterator).second.dwContext, CTX(m_eCtx)))
		{
			return (*theIterator).second.pValue;
		}
		else
		{
			return csText;
		}
	}
	
	// now try the user map
	theIterator = m_userVarMap.find(csUpperVar);
	
	// did we get one?
	if (theIterator!=m_userVarMap.end())
	{
		// return its value
		return (*theIterator).second.pValue;
	}
	
	// not found
	return csText;
}


//////////////////////////////////////////////////////////////////////////
// is csOper an operator?

bool CSAScriptEval::GetOperator(const CSAString & csOper, CSAScrOperator & oper) const
{
	// operator names are stored in the map as upper case
	CSAString csUpperOper = csOper;
	csUpperOper.MakeUpper();
	
	// find that operator in the pre-defined operator map
	CSAScrOperatorMap::const_iterator theIterator;
	theIterator = g_varOperManager.m_operatorMap.find(csUpperOper);
	
	bool ok = false;
	
	// did we get one?
	if (theIterator!=g_varOperManager.m_operatorMap.end())
	{
		ok = true;
		oper = (*theIterator).second;
	}
	else
	{
		// no? try the local map
		CSAScrOperatorMap::const_iterator theIterator;
		theIterator = m_localOperMap.find(csUpperOper);
		
		// did we get one?
		if (theIterator!=m_localOperMap.end())
		{
			ok = true;
			oper = (*theIterator).second;
		}
		else
		{
			ok = false;
		}
	}
	
	return ok;
}

//////////////////////////////////////////////////////////////////////////
// evaluate the operator

CSAString CSAScriptEval::DoOperator(const CSAScrOperator &oper, CSAScrOpStack & paramStack)
{
	CSAString csResult = SA_VAR_ERROR;
	
	// is it really a good one?
	if (oper.m_pFnOperator!=NULL)
	{
		// right number of params?
		
		//
		//	[JRT]	_T( "signed/unsigned" ) Mismatch...
		//
//		if (oper.m_iParamCount <= paramStack.size())
		if (oper.m_iParamCount <= (int)paramStack.size())
		{
			// ok, let's go
			csResult = ((pVarEvalFunc)(oper.m_pFnOperator))(paramStack, oper);
			
			// special recursion for Import
			if (oper.m_pFnOperator==__operImport)
			{
				CSAScriptEval t(m_localVarMap, m_userVarMap);
				csResult = t.ParseString(csResult, (eParseContext)CTXALL);
			}
			// special handling for VarSet
			else if (oper.m_pFnOperator==__operVarSet)
			{
				CSAString csVal = PopText(paramStack);
				CSAString csVar = PopText(paramStack);
				m_userVarMap.Add(csVar, csVal, CTXALL);
			}
			
		}
		else
		{
			//assert(0);
			throw CSAScrException(CSAScrException::eParseInternal, _T( "Not enough operands on the stack" ));
		}
	}
	
	return csResult;
}

//////////////////////////////////////////////////////////////////////////
// expand a macro expression. formal parameters are replaced 
// by input parameters.
// ex. &_max(Div(a, b), c) => If(Gtr(Div(a, b), c), Div(a, b), c)

void CSAScriptEval::ExpandMacro(const CSAScrOperator & macroOper, 
                                CTokenizer &tokenizer, 
                                int iStartToken, 
                                int iEndToken, 
                                CSAScrOpStack &outExpStack)
{
	assert(macroOper.m_bMacroFunction);
	
	////////////////////////////////////////////////////
	// 1. break the arg list into individual args
	CSAStringArray csaFormalParams; 
	CSAMacro::GetFormalParamArray(macroOper.m_csParamName, csaFormalParams);
	
	////////////////////////////////////////////////////
	// 2. get the input parameters.
	// these may be expressions themselves...
	CSAStringArray csaInputParams;
	GetParamArray(tokenizer, iStartToken + 2, iEndToken - 1, csaInputParams);
	
	// check for mismatched param counts
	if (csaFormalParams.GetSize() != csaInputParams.GetSize())
	{
		//assert(0);
		CSAString t;
		t.Format(_T( "Operator %s expects %d parameters" ), macroOper.m_csOperatorName, macroOper.m_iParamCount);
		throw CSAScrException(CSAScrException::eParseParameterCount, t);
	}
	
	////////////////////////////////////////////////////
	// 3. replace formal params with input params
	CSAString csFunc = macroOper.m_csFunctionString;
	
	// tokenize the function
	CTokenizer funcTokenizer(NULL, 0, m_MetaChars, m_iMetaChars);
	if (funcTokenizer.Tokenize(csFunc)!=CTokenizer::eErrorNone)
	{
		throw CSAScrException(CSAScrException::eParseSyntax, funcTokenizer.GetLastToken().m_csToken);
	}
	
	csFunc = _T( "" );
	
	int iFuncTokens = funcTokenizer.GetTokenCount();
	for (int i=0; i < iFuncTokens; i++)
	{
		CSAToken token;
		token = funcTokenizer.GetToken(i);
		
#ifdef _DEBUG
		if (i==0)
		{
			assert(token.m_csToken==g_variablePrefix);
		}
#endif
		if (token.m_csToken == g_variablePrefix)
		{
			assert(i==0);
			continue;
		}
		
		if (token.m_bIsQuotedString)
		{
			csFunc+=_T( "\"" );
			csFunc+=token.m_csToken;
			csFunc+=_T( "\"" );
			continue;
		}
		
		
		// is this token a formal parameter ?
		for (int z=0; z < csaFormalParams.GetSize(); z++)
		{
			if (token.m_csToken == csaFormalParams.GetAt(z))
			{
				// yes. replace it with the corresponding input parameter
				token.m_csToken = csaInputParams.GetAt(z);
				break;
			}
		}
		
		// rebuild the function string
		csFunc+=token.m_csToken;
	}
	
#ifndef SASCR_SILENT
	SAScr::SAScrTrace(_T( "Macro %s expanded: %s ==> %s\n" ), macroOper.m_csOperatorName, macroOper.m_csFunctionString, csFunc);
#endif
	
	////////////////////////////////////////////////////
	// 4. tokenize the new function... whew
	funcTokenizer.Init();
	if (funcTokenizer.Tokenize(csFunc)!=CTokenizer::eErrorNone)
	{
		throw CSAScrException(CSAScrException::eParseSyntax, funcTokenizer.GetLastToken().m_csToken);
	}
	
	////////////////////////////////////////////////////
	// 5. put the new expression on the stack. note that this is recursive.
	ExpressionToStack(outExpStack, funcTokenizer, 0, funcTokenizer.GetTokenCount());
}

//////////////////////////////////////////////////////////////////////////
// decompose the input expression to find the input parameters.

void CSAScriptEval::GetParamArray(CTokenizer &tokenizer, int iStartToken, int iEndToken, CSAStringArray &csaParams)
{
   assert(iStartToken<=iEndToken);

   //enum {eStateWantComma, eStateWantCloseParen

   CSAString csCurParam;
   int iParenLevel = 0;
   int iCommaCount = 0;
   bool bHaveParamToAdd = false;

   for (int i=iStartToken;i<iEndToken;i++)
   {
     CSAToken token;
     token = tokenizer.GetToken(i);
  
     if (token.m_bIsQuotedString)
     {
        // re-quote
        token.m_csToken = _T( "\"" ) + token.m_csToken + _T( "\"" );
     }
  
     // if we find a comma in this scope (not down in sub paren levels)
     if ((token.m_csToken == g_operParamSeparator) && (iParenLevel == 0))
     {
        iCommaCount++;
        csaParams.Add(csCurParam);
        csCurParam = _T( "" );
        bHaveParamToAdd = false;
        continue;
     }
  
     csCurParam+=token.m_csToken;      
  
     bHaveParamToAdd = true; 
  
     // starting sub-expression
     if (token.m_csToken == g_operSeparatorStart)
     {
        iParenLevel++;
     
        // store what we already have
        if (iParenLevel==0)
        {
           csaParams.Add(csCurParam);
           assert(iCommaCount==csaParams.GetSize() - 1);
           csCurParam = _T( "" );
           bHaveParamToAdd = false;
        }
     
        continue;
     }
  
     // close paren. 
     if (token.m_csToken == g_operSeparatorStop)
     {
        iParenLevel--;
     
        // is this the end of the sub expression?
        if (iParenLevel==0)
        {
           // yes.
        
           csaParams.Add(csCurParam);
           assert(iCommaCount==csaParams.GetSize() - 1);
           csCurParam = _T( "" );
           bHaveParamToAdd = false;
        
           // the next token must be a comma?
           i++;
           if (i<iEndToken)
           {
              token = tokenizer.GetToken(i);
              if (token.m_csToken == g_operParamSeparator)
              {
                 // end of sub expression
                 iCommaCount++;
                 continue;
              }
              else 
              {
                 throw CSAScrException(CSAScrException::eParseSyntax, token.m_csToken);
              }
           }
           else if (i==iEndToken)
           {
              // end of parameter list . we must be done
              continue;
           }
           else
           {
              throw CSAScrException(CSAScrException::eParseSyntax, token.m_csToken);
           }
        }
     
        continue;
     }
   }

   if (bHaveParamToAdd)
   {
     csaParams.Add(csCurParam);
   }

   assert(iCommaCount==csaParams.GetSize() - 1);
}


//////////////////////////////////////////////////////////////////////////

void CSAScriptEval::SetLocalOperators(const CSAScrOperatorMap &localOperatorMap)
{
	m_localOperMap = localOperatorMap;   
}


////////////////////////////////////////////////////////////////

void CSAScriptEval::ParseMacroFromString(const TCHAR	*pString, CSAMacroMap &macros)
{
	CTokenizer macroTokenizer(NULL, 0, m_MetaChars, m_iMetaChars);
	
	if (macroTokenizer.Tokenize(pString)!=CTokenizer::eErrorNone)
	{
		throw CSAScrException(CSAScrException::eParseSyntax, macroTokenizer.GetLastToken().m_csToken);
	}
	
	int iTokens = macroTokenizer.GetTokenCount();
	
	/* 
	macro defs must follow this form:
	
      Macro: funcName(params) &expression
	  
		ex:
		Macro: MyFunc(__p1, __p2) &Join(Upper(__p1), Lower(__p2))
	*/
	
	enum {eStateWantDefine, 
		eStateWantFuncName, 
		eStateWantParamOpenParen, 
		eStateWantParams, 
		eStateWantFuncStart,
		eStateReadingFunction};
	
	int iState = eStateWantDefine;
	
	int iCurToken = 0;
	//
	//	[JRT]	This Is Not Used...?  Feature Or Flaw?
	//
//	bool bOK = true;
	CSAString csAccum;
	int iParenAccum = 0;
	
	CSAMacro curMacro;
	
	while (iCurToken < iTokens)
	{
		
		CSAToken curToken = macroTokenizer.GetToken(iCurToken);
		
		switch (iState)
		{
		case eStateWantDefine:
			if (curToken.m_csToken == MACRO_DEFINITION_START)
			{
				iState = eStateWantFuncName;
			}
			else
			{
				throw CSAScrException(CSAScrException::eParseSyntax, macroTokenizer.GetLastToken().m_csToken);
			}
			break;
		case eStateWantFuncName:
			curMacro.m_csName = curToken.m_csToken;
			iState = eStateWantParamOpenParen;
			break;
		case eStateWantParamOpenParen:
			if (curToken.m_csToken == g_operSeparatorStart)
			{
				iState = eStateWantParams;
			}
			else
			{
				throw CSAScrException(CSAScrException::eParseSyntax, macroTokenizer.GetLastToken().m_csToken);
			}
			break;
		case eStateWantParams:
			if (curToken.m_csToken == g_operSeparatorStop)
			{
				// set this as the param list
				curMacro.m_csParams = csAccum;
				csAccum = _T( "" );
				iState = eStateWantFuncStart;
			}
			else
			{
				// add this to the param list
				csAccum+=curToken.m_csToken;
			}
			break;
		case eStateWantFuncStart:
			if (curToken.m_csToken == g_variablePrefix)
			{
				iState = eStateReadingFunction;
				iParenAccum = 0;
				
				// get up to first open paren
				for (;iCurToken < iTokens;iCurToken++)
				{
					curToken = macroTokenizer.GetToken(iCurToken);
					
					// add this to the function
					csAccum+=curToken.m_csToken;
					
					if (curToken.m_csToken == g_operSeparatorStart)
					{
						iParenAccum++;
						break;
					}
				}
				
				if (iParenAccum==0) // didn't find an open paren??
				{
					throw CSAScrException(CSAScrException::eParseSyntax, macroTokenizer.GetLastToken().m_csToken);
				}
			}
			else
			{
				throw CSAScrException(CSAScrException::eParseSyntax, macroTokenizer.GetLastToken().m_csToken);
			}
			break;
		case eStateReadingFunction:
			if (curToken.m_bIsQuotedString)
			{
				// re-quote quoted text
				csAccum+=_T( '\"' );
				csAccum+=curToken.m_csToken;
				csAccum+=_T( '\"' );
			}
			else
			{
				csAccum+=curToken.m_csToken;
			}
			
			if (curToken.m_csToken == g_operSeparatorStart)
			{
				iParenAccum++;
			}
			
			if (curToken.m_csToken == g_operSeparatorStop)
			{
				iParenAccum--;
			}
			
			// done
			if (iParenAccum==0) 
			{
				curMacro.m_csFunc = csAccum;
				
				macros.Add(curMacro);
				
				curMacro.Clear();
				csAccum = _T( "" );
				iParenAccum = 0;
				iState = eStateWantDefine;
			}
			
			break;
      }
	  
      iCurToken++;
	  
   }
}


