/*********************************************************************

   Copyright (C) 2002 Smaller Animals Software, Inc.

   This software is provided 'as-is', without any express or implied
   warranty.  In no event will the authors be held liable for any damages
   arising from the use of this software.

   Permission is granted to anyone to use this software for any purpose,
   including commercial applications, and to alter it and redistribute it
   freely, subject to the following restrictions:

   1. The origin of this software must not be misrepresented; you must not
     claim that you wrote the original software. If you use this software
     in a product, an acknowledgment in the product documentation would be
     appreciated but is not required.

   2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.

   3. This notice may not be removed or altered from any source distribution.

   http://www.smalleranimals.com
   smallest@smalleranimals.com

**********************************************************************/

#ifndef VARSH
#define VARSH

#include "SAScrDefs.h"
#include "SAScrString.h"
#include "SAScrStringArray.h"
#include "SAScrOper.h"
#include "SAScrVariable.h"
#include "SAScrTokenizer.h"
#include "SAScrOperManager.h"


class CSAScriptEval
{
public:			 
   ////////////////////////////////////////////////////////////////
   // these are the two calls you'll need most often.

   // init with local variables
	CSAScriptEval(const CSAScrVariableMap &vars);

   // main entry point. 
   CSAString  ParseString(const TCHAR * pText, eParseContext eCtx);
   
   ////////////////////////////////////////////////////////////////
   //
   // advanced operation

   /****************************************************************** 
      parse a string into macros

      the string may contain more than one macro.

      macro definitions in the string must follow this form:

      Macro: funcName(params) &expression

      ex:
      Macro: MyFunc(__p1, __p2) &Join(Upper(__p1), Lower(__p2))
      Macro: MyFunc2(__p1) &Upper(__p1)

      This is a good way to let users define their own macros at run-time.
      To get the macros into the evaluator:

         CSAScriptEval eval(vars);

         // turn the string into one or more macros
         CSAMacroMap macroMap;
         eval.ParseMacroFromString(csMacroDefinitions, macroMap);

         // create operators from the macros
         CSAScrOperatorMap operMap;
         operMap.AddMacros(macroMap);

         // tell the evaluator to use these operators (which were created from the macros)
         eval.SetLocalOperators(operMap);

         eval.ParseString(...);

   ******************************************************************/
   static void ParseMacroFromString(const TCHAR *pString, CSAMacroMap &macros);

   // in case you want to define some temporary operators (usually local macros)
   void  SetLocalOperators(const CSAScrOperatorMap &localOperatorMap);

   /******************************************************************
      FindExpressions 
      This can be used to find a COLORREF vector for the string.
      Specifically, this function was written to allow a syntax colorizing 
      text editor to get colors for an expression. The editor is not
      included here, but you might find it useful to see how this function
      breaks up expressions into various pieces.

      Use is just like ParseString
   ******************************************************************/
   void  FindExpressions(vector<COLORREF> & vOutColors, const CSAString &csText, COLORREF clrNon, COLORREF clrExp, COLORREF clrQot, COLORREF clrVar, COLORREF clrOper, COLORREF clrErr) const;

protected:
   // the special TCHARacters
   static TCHAR m_MetaChars[];
   static int  m_iMetaChars;

protected:
   //protected constructor used when we need to spawn a secondary parser for a macro or an import
   CSAScriptEval(const CSAScrVariableMap &localVars, 
                  const CSAScrVariableMap &userVars);
   
   // does this look like an expression?
	bool		   IsExpression(CTokenizer &tokenizer, int iStartToken) const;

   // is this a variable?
   bool        IsVariable(const CSAString& csText) const;
	
   // find this operator in our maps
   bool        GetOperator(const CSAString & csOper, CSAScrOperator &oper) const;

   // execute this operator
	CSAString	DoOperator(const CSAScrOperator &oper, CSAScrOpStack &strParams);

   // stack-ize the expression in m_tokenizer from iStartToken to iEndToken. 
   void		   ExpressionToStack(CSAScrOpStack &expStack, CTokenizer &tokenizer, int iStartToken, int iEndToken);

   // main evaluation loop
   CSAString	Evaluate(int iStartToken, int iEndToken);

   // do a lookup on this variable
   CSAString	ParseVariable(const CSAString &csText);

   // get a token
	CSAString	GetToken(const CSAString &csCur);

   // find the index of the token (in m_tokenizer) that ends this expression
   int         FindLastToken(CTokenizer &tokenizer, int iStartToken) const;

   // run through the string, evaluating expressions
   void		   ParseStringA(CSAString & csOut, const CSAString &csText);

   // reaplce formal parameters with their passed-in equivalents, tokenize the whole thing
   void        ExpandMacro(const CSAScrOperator & macroOper, CTokenizer &tokenizer, int iStartToken, int iEndToken, CSAScrOpStack &outExpStack);

   // push variable state
   void        PushState(CSAScrOpStack & paramStack);

   // replace macro locals with real params
   CSAString   PTokenReplace(const CSAString &csFunc, CSAStringArray &csaParams, CSAStringArray &csaPTokens);

   // get the array of parameters for a macro
   void        GetParamArray(CTokenizer &tokenizer, int iStartToken, int iEndToken, CSAStringArray &csaParams);

protected:
   // parse context
	eParseContext  m_eCtx;

   // local variables
   CSAScrVariableMap    m_localVarMap;

   // user variables
   CSAScrVariableMap    m_userVarMap;

   // recursion count
	int			   m_iRefCount;

   // handy collection string
   CSAString     m_csSepString;

   // stores our tokenized lines
   CTokenizer      m_tokenizer;

   // local operator map
   CSAScrOperatorMap m_localOperMap;
   
   // recursion limit
   enum {VARS_MAX_REF=10};
};

#endif