/*********************************************************************

   Copyright (C) 2002 Smaller Animals Software, Inc.

   This software is provided 'as-is', without any express or implied
   warranty.  In no event will the authors be held liable for any damages
   arising from the use of this software.

   Permission is granted to anyone to use this software for any purpose,
   including commercial applications, and to alter it and redistribute it
   freely, subject to the following restrictions:

   1. The origin of this software must not be misrepresented; you must not
     claim that you wrote the original software. If you use this software
     in a product, an acknowledgment in the product documentation would be
     appreciated but is not required.

   2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.

   3. This notice may not be removed or altered from any source distribution.

   http://www.smalleranimals.com
   smallest@smalleranimals.com

**********************************************************************/
/********************************************************

   VarEvalFuncsEx

   These are extensions to the core set of built-in SAScript 
   operators.

   If you want to add more built-in operators, this is a 
   good place to add them.


  Here's the general form for a built-in operator function:

   // &MyOperator(p1, p2, p3)
   CSAString __operYourOperatorFunction(CSAScrOpStack &paramStack, const CSAScrOperator &oper)
   {
      // default error string
	   CSAString csResult = SA_VAR_ERROR;

      // pop the parameters off. 
      // because this is a stack, parameter order is revered
   
      CSAString csParam3 = PopText(paramStack);
      CSAString csParam2 = PopText(paramStack);
      CSAString csParam1 = PopText(paramStack);

      csResult = ...;

      return csResult;
   }

   How to add your own operators:

   1. Create a function like the one above

   2. Declare your function in SASrcEvalFuncsEx.h, like this:

      extern VarEvalFunc __operYourOperatorFunction;

   3. Create a const TCHAR	* representation of your operator name
      in this file (see below), like this :
      
        const TCHAR	* g_operYourOperatorName = _T( "MyOperator" );

      This is the name users will use to call your operator.

   4. Add an entry to the g_operArrayEx array , like this (see this file, below):

      CSAScrOperator(g_operYourOperatorName,    // operator name
                     __operYourOperatorFunction,// pointer to the function
                     3,                         // param count
                     _T( "p1, p2, p3" ) ,             // parameter list
                     _T( "" ),                        // leave blank
                     _T( "File size, in kilobytes" )),// function description text
      
********************************************************/


#include "stdafx.h" 
#include "SAScrDefs.h" 
#include "SAScrUtils.h" 
#include "SAScrEvalFuncsEx.h"
#include "SAScrOperManager.h"

#include <math.h>
#include <time.h>
#include <direct.h>
#include <fcntl.h>
#include <io.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/utime.h>


//////////////////////////////////////////////////////////////////////////
// names for the extension operators

const TCHAR	* g_operCurDate		= _T( "SysDate" );
const TCHAR	* g_operCurTime		= _T( "SysTime" );

const TCHAR	* g_operFileSizeB	   = _T( "Size" );
const TCHAR	* g_operFileSizeK	   = _T( "SizeK" );
const TCHAR	* g_operDate			= _T( "Date" );
const TCHAR	* g_operTime			= _T( "Time" );
const TCHAR	* g_operFileExist    = _T( "FExist" );

const TCHAR	* g_operConcat3		= _T( "Join3" );
const TCHAR	* g_operQuote		   = _T( "Q" );

//////////////////////////////////////////////////////////////////////////
// name, function, param count, param name string, function name, function description

CSAScrOperator g_operArrayEx[] = 
{
	CSAScrOperator(g_operCurDate,	__operCurDateString,	   0 ,_T( "" ) ,        _T( "" ), _T( "Current system date" )),
	CSAScrOperator(g_operCurTime,	__operCurTimeString,	   0 ,_T( "" ) ,        _T( "" ), _T( "Current system time" )),
	CSAScrOperator(g_operFileSizeB,	__operFileSizeBString,	1 ,_T( "Path" ) , _T( "" ), _T( "File size, in bytes" )),
	CSAScrOperator(g_operFileSizeK,	__operFileSizeKString,	1 ,_T( "Path" ) , _T( "" ), _T( "File size, in kilobytes" )),
	CSAScrOperator(g_operDate,		__operFileDateString,	1 ,_T( "Path" ) ,    _T( "" ), _T( "File modification date" )),
	CSAScrOperator(g_operTime,		__operFileTimeString,	1 ,_T( "Path" ) ,    _T( "" ), _T( "File modification time" )),
	CSAScrOperator(g_operConcat3,	__operConcat3,			   3 ,_T( "A, B, C" ),  _T( "" ), _T( "Concatenate A, B and C to ABC" )),
	CSAScrOperator(g_operQuote,		__operQuoteStr,		1 ,_T( "Text" ) ,    _T( "" ), _T( "Output text in double quotes" )),
	CSAScrOperator(g_operFileExist,	__operFileExists,		1 ,_T( "file" ) ,    _T( "" ), _T( "If \"file\" exists output \"1\", else \"0\"") ),
};

//////////////////////////////////////////////////////////////////////////
// called by InitOperatorMap to set up addtional built-in operators 

void CSAScrOpManager::InitOperatorMapEx(CSAScrOperatorMap& operMap)
{
   int iOpers = ARRAYSIZE(g_operArrayEx);

   for (int i=0; i < iOpers; i++)
   {
      operMap.Add(g_operArrayEx[i]);
   }
}

//////////////////////////////////////////////////////////////////////////

// start add-in operators

//////////////////////////////////////////////////////////////////////////

CSAString __operCurTimeString(CSAScrOpStack& /*paramStack*/, const CSAScrOperator& /*oper*/)
{
	CSAString csResult = SA_VAR_ERROR;

   csResult = SAScrEx::GetCurTimeString();

   return csResult;
}

//////////////////////////////////////////////////////////////////////////

CSAString __operCurDateString(CSAScrOpStack& /*paramStack*/, const CSAScrOperator & /*oper*/)
{
	CSAString csResult = SA_VAR_ERROR;

   csResult = SAScrEx::GetCurDateString();

   return csResult;
}

//////////////////////////////////////////////////////////////////////////

CSAString __operFileDateString(CSAScrOpStack &paramStack, const CSAScrOperator& /*oper*/)
{
	CSAString csResult = SA_VAR_ERROR;

	CSAString csParam1 = PopText(paramStack);
   

	csResult = SAScrEx::GetFileDateOnlyString(csParam1, true);

   return csResult;
}

//////////////////////////////////////////////////////////////////////////

CSAString __operFileTimeString(CSAScrOpStack &paramStack, const CSAScrOperator& /*oper*/)
{
	CSAString csResult = SA_VAR_ERROR;

	CSAString csParam1 = PopText(paramStack);
   

   csResult = SAScrEx::GetFileTimeOnlyString(csParam1, true);

   return csResult;
}

//////////////////////////////////////////////////////////////////////////

CSAString __operFileSizeBString(CSAScrOpStack &paramStack, const CSAScrOperator& /*oper*/)
{
	CSAString csResult = SA_VAR_ERROR;

	CSAString csParam1 = PopText(paramStack);
   

   csResult = SAScrEx::GetFileSizeStringB(csParam1);

   return csResult;
}

//////////////////////////////////////////////////////////////////////////

CSAString __operFileSizeKString(CSAScrOpStack &paramStack, const CSAScrOperator& /*oper*/)
{
	CSAString csResult = SA_VAR_ERROR;

	CSAString csParam1 = PopText(paramStack);
   

   csResult = SAScrEx::GetFileSizeString(csParam1);

   return csResult;
}

//
// end built-in operators
//
//////////////////////////////////////////////////////////////////////
//
// start utility helper functions
//

CSAString SAScrEx::GetCurTimeString()
{
   struct tm *today;
   _tzset();

   time_t ltime;
   time(&ltime);
   
   today = localtime( &ltime );

   TCHAR buf[256];
   _tcsftime(buf, 256, _T( "%H:%M:%S" ), today);
	
   CSAString csTime = buf;

	return csTime;
}

//////////////////////////////////////////////////////////////////////

CSAString SAScrEx::GetCurDateString()
{
  struct tm *today;
   _tzset();

   time_t ltime;
   time(&ltime);
   
   today = localtime( &ltime );

   TCHAR buf[256];
   _tcsftime(buf, 256, _T( "%b %d %Y" ), today);
	
   CSAString csTime = buf;

	return csTime;
}

//////////////////////////////////////////////////////////////////////

CSAString SAScrEx::DateString(time_t t)
{
 	struct tm* ltime;
	ltime = localtime(&t);

	TCHAR buf[256];
	_tcsftime(buf, 256, _T( "%m/%d/%y %H:%M:%S" ), ltime);

	return buf;
}


CSAString SAScrEx::GetFileSizeStringB(const CSAString &csFQSrcPath)
{
	struct _stat buf;
	int result = _tstat( csFQSrcPath, &buf );
	if( result != 0 ) 
   {
      CSAString csResult;
      csResult.Format(_T( "<!-- Can't open file: %s -->" ), csFQSrcPath);
      return csResult;
	}

	CSAString csSize;
	csSize.Format(_T( "%u" ), buf.st_size);

	return csSize;
}

//////////////////////////////////////////////////////////////////////

CSAString SAScrEx::GetFileSizeString(const CSAString &csFQSrcPath)
{
	struct _stat buf;
	int result = _tstat( csFQSrcPath, &buf );
	if( result != 0 ) 
	{
      CSAString csResult;
      csResult.Format(_T( "<!-- Can't open file: %s -->" ), csFQSrcPath);
      return csResult;

	}

	double size = (double)buf.st_size; // size

	CSAString csSize;
	csSize.Format(_T( "%2.1f" ), size / 1024.0);

	return csSize;
}

//////////////////////////////////////////////////////////////////////

CSAString SAScrEx::GetFileDateString(const CSAString &csFQSrcPath, bool bMod)
{
	struct _stat buf;
	int result = _tstat( csFQSrcPath, &buf );
	if ( result != 0 ) 
   {
      CSAString csResult;
      csResult.Format(_T( "<!-- Can't open file: %s -->" ), csFQSrcPath);
      return csResult;

	}

	time_t ft = buf.st_ctime; // creation time
	if (bMod)
		ft = buf.st_mtime; // mod time

   struct tm *today;
   today = localtime( &ft );

   TCHAR cbuf[256];
	_tcsftime(cbuf, 256, _T( "%c" ), today);

	CSAString csDate;
	csDate = cbuf;

	return csDate;
}

//////////////////////////////////////////////////////////////////////

CSAString SAScrEx::GetFileDateOnlyString(const CSAString &csFQSrcPath, bool bMod)
{
	struct _stat buf;
	int result = _tstat( csFQSrcPath, &buf );
	if ( result != 0 ) 
   {
      CSAString csResult;
      csResult.Format(_T( "<!-- Can't open file: %s -->" ), csFQSrcPath);
      return csResult;
	}

	time_t ft = buf.st_ctime; // creation time
	if (bMod)
		ft = buf.st_mtime; // mod time

   struct tm *today;
   today = localtime( &ft );

   TCHAR cbuf[256];
	_tcsftime(cbuf, 256, _T( "%b %d %Y" ), today);

	CSAString csDate;
	csDate = cbuf;


	return csDate;
}

//////////////////////////////////////////////////////////////////////

CSAString SAScrEx::GetFileTimeOnlyString(const CSAString &csFQSrcPath, bool bMod)
{
	struct _stat buf;
	int result = _tstat( csFQSrcPath, &buf );
	if ( result != 0 ) 
   {
      CSAString csResult;
      csResult.Format(_T( "<!-- Can't open file: %s -->" ), csFQSrcPath);
      return csResult;
	}

	time_t ft = buf.st_ctime; // creation time
	if (bMod)
		ft = buf.st_mtime; // mod time

   struct tm *today;
   today = localtime( &ft );

   TCHAR cbuf[256];
	_tcsftime(cbuf, 256, _T( "%H:%M:%S" ), today);

	CSAString csTime;
	csTime = cbuf;

	return csTime;
}

/////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////

CSAString __operQuoteStr(CSAScrOpStack &paramStack, const CSAScrOperator& /*oper*/)
{
	CSAString csResult = SA_VAR_ERROR;

	CSAString csParam = PopText(paramStack);

   csResult = _T( "\" )_T( " + csParam + " )\"" );

   return csResult;
}      

//////////////////////////////////////////////////////////////////////////

CSAString __operConcat3(CSAScrOpStack &paramStack, const CSAScrOperator& /*oper*/)
{
	CSAString csResult = SA_VAR_ERROR;

   CSAString csParam3 = PopText(paramStack);

   CSAString csParam2 = PopText(paramStack);
   
   CSAString csParam1 = PopText(paramStack);
   

   csResult = csParam1 + csParam2 + csParam3;

   return csResult;
}

//////////////////////////////////////////////////////////////////////////

CSAString __operFileExists(CSAScrOpStack &paramStack, const CSAScrOperator& /*oper*/)
{
	CSAString csResult = SA_VAR_ERROR;

   CSAString csParam1 = PopText(paramStack);

   if (_taccess(csParam1, 0)==0)
   {
      csResult = SA_VAR_TRUE;
   }
   else
   {
      csResult = SA_VAR_FALSE;
   }

   return csResult;
}