/*********************************************************************

   Copyright (C) 2002 Smaller Animals Software, Inc.

   This software is provided 'as-is', without any express or implied
   warranty.  In no event will the authors be held liable for any damages
   arising from the use of this software.

   Permission is granted to anyone to use this software for any purpose,
   including commercial applications, and to alter it and redistribute it
   freely, subject to the following restrictions:

   1. The origin of this software must not be misrepresented; you must not
     claim that you wrote the original software. If you use this software
     in a product, an acknowledgment in the product documentation would be
     appreciated but is not required.

   2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.

   3. This notice may not be removed or altered from any source distribution.

   http://www.smalleranimals.com
   smallest@smalleranimals.com

**********************************************************************/

#include "stdafx.h"
#include "SAScrTokenizer.h"
#include "SAScrUtils.h"
#include "SAScrException.h"

// illegal escape, etc..
#define SASCRTOKENIZER_THROW_EXCEPTIONS

// SAScript does not allow comments!!
#undef ALLOW_COMMENTS

#ifdef ALLOW_COMMENTS
#define COMMENT_START   _T( "<!" )
#define COMMENT_STOP    _T( "!>" )
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CTokenizer::CTokenizer(TCHAR	**pKeyWords, 
                       int iKeyWords,
                       TCHAR	*pMetaChars,
                       int iMetaChars)
{
   m_pKeyWords = pKeyWords;
   m_iKeyWords = iKeyWords;
   m_MetaChars = pMetaChars;
   m_iMetaChars = iMetaChars;

   m_bNoTranslateEscapes = true;

}

//////////////////////////////////////////////////////////////////////
// bye bye
CTokenizer::~CTokenizer()
{

}

//////////////////////////////////////////////////////////////////////
// remove tokens
void CTokenizer::Clear()
{
   m_tokens.clear();
}

void CTokenizer::Init()
{
   Clear();
}

//////////////////////////////////////////////////////////////////////
// extract an array of tokens from this input text
int CTokenizer::Tokenize(const TCHAR	*pInputLine)
{
   int err = eErrorNone;

   Init();

   int iLen = _tcslen(pInputLine);

   bool bInComment = false;

   for (int i=0; i < iLen; i++)
   {
      CSAString curTokenString;
      bool bQuotedString = false;

      int iConsumed = GetToken(pInputLine + i, curTokenString, bQuotedString);

      if (iConsumed > 0)
      {
         //if (curTokenString.length() > 0)
         {
#ifdef ALLOW_COMMENTS
            if (curTokenString == COMMENT_START)
            {
               bInComment = true;
               i+=iConsumed;
               continue;
            }
            
            if (curTokenString == COMMENT_STOP)
            {
               if (!bInComment)
               {
                  err = eErrorSyntax;
                  break;
               }

               bInComment = false;

               i+=iConsumed;
               continue;
            }
#endif

            if (!bInComment)
            {
               int iStart = i;
               int iStop = i + iConsumed - 1;

               CSAToken curToken(curTokenString, iStart, iStop, bQuotedString);

               m_tokens.push_back(curToken);
            }
         }
         
         i+=iConsumed;

         i--; // back up! the for iteration will increase i
         continue;
      }
      else
      {
         if (IsMetaChar(*(pInputLine + i)))
         {
            if (!_istspace(*(pInputLine + i)))
            {
               curTokenString = _T( "" );
               curTokenString+=*(pInputLine + i);

               CSAToken curToken(curTokenString, i, curTokenString.GetLength() - 1, bQuotedString);

               m_tokens.push_back(curToken);
            }
         }
      }
   }

   return err;
}

//////////////////////////////////////////////////////////////////////
// is this text a keyword?
bool CTokenizer::IsKeyWord(CSAString &str)
{
   return IsKeyWord((const TCHAR	*)str);
}

//////////////////////////////////////////////////////////////////////

bool CTokenizer::IsKeyWord(const TCHAR	*pInput)
{
   if (m_pKeyWords==NULL)
   {
      return false;
   }

   int i_tcslen = _tcslen(pInput);

   int iBestMatch = -1;

   for (int i=0; i < m_iKeyWords; i++)
   {
      int iCurKWLen = _tcslen(m_pKeyWords[i]);

      if (i_tcslen <= iCurKWLen)
      {
         if (sa_strnicmp(m_pKeyWords[i], pInput, iCurKWLen)==0)
         {
            iBestMatch = i;
         }
      }
   }

   if (iBestMatch==-1)
   {
#ifdef ALLOW_COMMENTS
      if (sa_strnicmp(COMMENT_START, pInput, _tcslen(COMMENT_START))==0)
      {
         iBestMatch = m_iKeyWords + 1;
      }

      if (sa_strnicmp(COMMENT_STOP, pInput, _tcslen(COMMENT_STOP))==0)
      {
         iBestMatch = m_iKeyWords + 2;
      }
#endif
   }

   return (iBestMatch != -1);
}

//////////////////////////////////////////////////////////////////////
// find the next token in the string

int CTokenizer::GetToken(const TCHAR	*pInput, CSAString &out, bool &bQuotedString)
{
   int i_tcslen = _tcslen(pInput);

   bool bFoundChars = false;

   bool bScanningMetaKW = false;

   bool bInQuotes = false;

   bQuotedString = false;

   TCHAR c;

   for (int iWS = 0; iWS < i_tcslen; iWS++)
   {
      c = *(pInput + iWS);
      if (!isspace(c))
         break;
      if (!iscntrl(c))
         break;
   }

   for (int i=iWS; i<i_tcslen; i++)
   {
      c = *(pInput + i);

      // not in quotes?
      // open quote
      if ((c==_T( '"' )) && (!bInQuotes))
      {
         // not in quotes, but we found a ", and we've already found chars?
         if (bFoundChars)
         {
            break;
         }

         bInQuotes = true;
#ifdef RETURN_QUOTED_STRINGS
         out+=c;
#endif
         continue;
      }

      // close quote
      if ((c==_T( '"' )) && (bInQuotes))
      {
         bInQuotes = false;
#ifdef RETURN_QUOTED_STRINGS
         out+=c;
#endif
         i+=1;

         bQuotedString = true;
         break;
      }

      // escapes in quotes
      if (bInQuotes)
      {
         // escape!
         if (c==_T( '\\' ))
         {
            if (i < i_tcslen - 1)
            {
               if (m_bNoTranslateEscapes)
               {
                  out+=c;
                  out+=*(pInput + i + 1);
                  i++;
                  continue;
               }
               else
               {
                  // translate the escape char
                  switch (*(pInput + i + 1))
                  {
                  case _T( '\\' ):
                     out+=_T( '\\' );
                     i++;
                     continue;
                     break;
                  case _T( '"' ):
                     out+=_T( '"' );
                     i++;
                     continue;
                     break;
#if 0 // not used by default
                  case _T( 'n' ):
                     out+=_T( '\n' );
                     i++;
                     continue;
                     break;
                  case _T( 'r' ):
                     out+=_T( '\r' );
                     i++;
                     continue;
                     break;
                  case _T( 't' ):
                     out+=_T( '\t' );
                     i++;
                     continue;
                     break;
#endif
                  default:
#ifdef SASCRTOKENIZER_THROW_EXCEPTIONS
                     {
                        CSAString t;
                        t.Format(_T( "Illegal escape sequence: \\%c" ), *(pInput + i + 1));
                        throw CSAScrException(CSAScrException::eParseSyntax, t);
                     }
#else
                     SAScr::SAScrTrace(_T( "Illegal escape \" )%c\_T( "\n" ), *(pInput + i + 1));
#endif
                     i++;
                     continue;
                     break;
                  }
               }
            }
         }

         out+=c;
         continue;
      }

      // break strings on ctrl or white space
      if (iscntrl(c) || isspace(c))
      {
         break;
      }

      // metachars...
      if (IsMetaChar(c))
      {
         if (iscntrl(c) || isspace(c))
         {
            break;
         }

         if (!bFoundChars)
         {
            bScanningMetaKW = true;

            bFoundChars = true;

            // in case some KWs start with metachars
				if (i < i_tcslen - 2)
				{
					TCHAR buf[3];
					buf[0]=c;
					buf[1]=*(pInput + i + 1);
					buf[2]=0;

					if (IsKeyWord(buf))
					{
						out+=c;
						out+=*(pInput + i + 1);

						i+=2;

						break;
					}
					else
					{
						buf[1] = 0;
						if (IsKeyWord(buf))
						{
							out+=c;

							i+=1;

							break;
						}
					}
				}

				out+=c;
         }
         else
         {
            break;
         }
      }
      else
      {
         if (!bFoundChars)
         {
            bScanningMetaKW = false;

            bFoundChars = true;

            out+=c;
         }
         else
         {
            if (bScanningMetaKW)
            {
               break;
            }
            else
            {
               out+=c;
            }
         }
      }
   }

   return i;
}

//////////////////////////////////////////////////////////////////////

bool CTokenizer::IsMetaChar(const TCHAR c)
{
   for (int i=0; i < m_iMetaChars; i++)
   {
      if (c==m_MetaChars[i])
      {

         return true;
      }
   }

   return false;
}

//////////////////////////////////////////////////////////////////////

bool CTokenizer::IsMetaChar(CSAString &str)
{
   if (str.GetLength() > 1)
   {
      return false;
   }

   TCHAR c = str.GetAt(0);

   for (int i=0; i < m_iMetaChars; i++)
   {
      if (c==m_MetaChars[i])
      {
         return true;
      }
   }

   return false;
}

//////////////////////////////////////////////////////////////////////

void CTokenizer::Dump()
{
   CSATokenVector::iterator theIterator;

   int i=0;
   for (theIterator = m_tokens.begin(); theIterator < m_tokens.end(); theIterator++)
   {
      SAScr::SAScrTrace(_T( "%d [%d-%d] : \"%s\"\n" ), i++, (*theIterator).m_iStart, (*theIterator).m_iStop, (const TCHAR	*)((*theIterator).m_csToken));
   }
}

//////////////////////////////////////////////////////////////////////

bool operator < (const CSAToken &a, const CSAToken &b)
{
   return (a.m_iStart < b.m_iStart);
}

bool operator == (const CSAToken &a, const CSAToken &b)
{
   return (a.m_iStart==b.m_iStart);
}

CSAString CTokenizer::TranslateEscapes(const CSAString &str)
{
   CSAString out;
   
   int i_tcslen = str.GetLength();
   const TCHAR	*pInput = (const TCHAR	*)str;

   for (int i=0; i < i_tcslen; i++)
   {
      TCHAR c = *(pInput + i);
      
      // escape!
      if (c==_T( '\\' ))
      {
         if (i < i_tcslen - 1)
         {
            // translate the escape char
            switch (*(pInput + i + 1))
            {
            case _T( '\\' ):
               out+=_T( '\\' );
               i++;
               continue;
               break;
            case _T( '"' ):
               out+=_T( '"' );
               i++;
               continue;
               break;
#if 0 // not in TN
            case _T( 'n' ):
               out+=_T( '\n' );
               i++;
               continue;
               break;
            case _T( 'r' ):
               out+=_T( '\r' );
               i++;
               continue;
               break;
            case _T( 't' ):
               out+=_T( '\t' );
               i++;
               continue;
               break;
#endif
            default:
#ifdef SASCRTOKENIZER_THROW_EXCEPTIONS
               {
                  CSAString t;
                  t.Format(_T( "Illegal escape sequence: \\%c" ), *(pInput + i + 1));
                  throw CSAScrException(CSAScrException::eParseSyntax, t);
               }
#else
               SAScr::SAScrTrace(_T( "Illegal escape \"%c\"\n" ), *(pInput + i + 1));
#endif
               i++;
               continue;
               break;
            }
         }
      }

      out+=c;
   }

   return out;
}